import os

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.helper_files import get_scr_components


class VarLicFilePrefix(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.licfileprefix',
                         severity='warning',
                         message='Prefix \'{PATH}\' to LIC_FILES_CHKSUM is not needed')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='LIC_FILES_CHKSUM')
        for i in items:
            for listitem in i.get_items():
                _comp = get_scr_components(listitem)
                _prefix = os.path.dirname(_comp['src'])
                if _prefix in ['${S}']:
                    res += self.finding(i.Origin, i.InFileLine,
                                        override_msg=self.Msg.format(PATH=_prefix))
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_scr_components
from oelint_parser.parser import INLINE_BLOCK


class VarSRCUriGitTag(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.srcurigittag',
                         severity='warning',
                         message='\'tag\' in SRC_URI-options leads to not-reproducible builds as git-tags can move around. Use explicit SRCREV')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')
        for item in items:
            if any([item.Flag.endswith(x) for x in ['md5sum', 'sha256sum']]):
                # These are just the hashes
                continue
            lines = [y.strip('\'') for y in item.get_items() if y]
            for x in lines:
                if x == INLINE_BLOCK:
                    continue
                _url = get_scr_components(x)
                if _url['scheme'] in ['git'] and 'tag' in _url['options']:
                    res += self.finding(item.Origin, item.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.constants import CONSTANTS


class VarSuggestedExists(Rule):
    def __init__(self):
        super().__init__(id='oelint.var.suggestedvar',
                         severity='info',
                         message='<FOO>',
                         onappend=False,
                         appendix=CONSTANTS.VariablesSuggested)

    def check(self, _file, stash):
        res = []
        _is_pkg_group = False
        for i in stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue='inherit'):
            if any(x == 'packagegroup' for x in i.get_items()):
                _is_pkg_group = True
                break
        for var in CONSTANTS.VariablesSuggested:
            if _is_pkg_group and var in ['LICENSE', 'CVE_PRODUCT']:
                continue
            items = stash.GetItemsFor(
                filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue=var)
            if not any(items):
                res += self.finding(_file, 0, 'Variable \'{var}\' should be set'.format(var=var), appendix=var)
        return res



import re

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.helper_files import expand_term
from oelint_parser.parser import INLINE_BLOCK


class VarImproperInherit(Rule):
    def __init__(self):
        super().__init__(id='oelint.var.improperinherit',
                         severity='error',
                         message='\'{INH}\' is not a proper bbclass name')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='inherit')
        for i in items:
            for subi in [expand_term(stash, _file, x) for x in i.get_items() if x and x != INLINE_BLOCK]:
                if not re.match(r'^[A-Za-z0-9_.-]+$', subi):
                    res += self.finding(i.Origin, i.InFileLine,
                                        self.Msg.replace('{INH}', subi))
        return res



import re
from oelint_parser.cls_item import Function
from oelint_adv.cls_rule import Rule


class TaskInstallNoCp(Rule):
    def __init__(self):
        super().__init__(id='oelint.task.heredocs',
                         severity='warning',
                         message='Usage of heredocs should be avoided. Use files instead')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER,
                                  attribute=Function.ATTR_FUNCNAME)
        for i in items:
            for index, line in enumerate(i.get_items()):
                line = line.strip()
                if re.match(r'^cat\s*>\s*.*<<\s*[A-Za-z0-9]+$', line) or re.match(r'^cat\s*<<\s*[A-Za-z0-9]+\s*>.*$', line):
                    res += self.finding(i.Origin, i.InFileLine + index)
        return res



from oelint_parser.constants import CONSTANTS

_NOINFO = False
_NOWARN = False
_RELPATH = False
_MESSAGEFORMAT = ''
_RULE_FILE = {}
_SUPPRESSIONS = []


def get_noinfo():
    return _NOINFO


def get_nowarn():
    return _NOWARN


def get_relpaths():
    return _RELPATH


def get_messageformat():
    return _MESSAGEFORMAT


def set_noinfo(value):
    global _NOINFO
    _NOINFO = value


def set_relpaths(value):
    global _RELPATH
    _RELPATH = value


def set_messageformat(value):
    global _MESSAGEFORMAT
    _MESSAGEFORMAT = value


def set_nowarn(value):
    global _NOWARN
    _NOWARN = value


def set_suppressions(value):
    global _SUPPRESSIONS
    _SUPPRESSIONS = value


def get_suppressions():
    return _SUPPRESSIONS


def get_rulefile():
    return _RULE_FILE


def set_rulefile(value):
    global _RULE_FILE
    _RULE_FILE = value
    CONSTANTS.AddFromRuleFile(value)



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarSRCUriGitTag(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.srcuriappend',
                         severity='error',
                         message='Use SRC_URI_append otherwise this will override weak defaults by inherit')

    def check(self, _file, stash):
        res = []
        inherits = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                     attribute=Variable.ATTR_VAR, attributeValue='inherit')
        if not inherits:
            return res

        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')
        for item in items:
            if any([item.Flag.endswith(x) for x in ['md5sum', 'sha256sum']]):
                # These are just the hashes
                continue
            if item.VarOp in [' += ']:
                res += self.finding(item.Origin, item.InFileLine)
        return res



import os
import re

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class FileNoSpaces(Rule):
    def __init__(self):
        super().__init__(id='oelint.file.underscores',
                         severity='error',
                         message='FOO',
                         onappend=False)

    def check(self, _file, stash):
        res = []
        _basename, _ext = os.path.splitext(os.path.basename(_file))
        if _ext in ['.bb']:  # pragma: no cover
            _image_install = any(stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                   attribute=Variable.ATTR_VAR, attributeValue='IMAGE_INSTALL'))
            _inherits = set()
            for inh in stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                         attribute=Variable.ATTR_VAR, attributeValue='inherit'):
                _inherits.update([x.strip() for x in re.split(
                    r'\s|,|\t|\x1b', inh.VarValue) if x])
            if _image_install or any(x in _inherits for x in ['core-image', 'image']):
                return res
            _sep = [x for x in _basename if x in ['_', '-']]
            _us = [x for x in _sep if x == '_']
            if len(_us) > 1:
                res += self.finding(_file, 1,
                                    override_msg='Filename should not contain more than one \'_\'')
            elif not _us or _sep[-1] != '_':
                res += self.finding(
                    _file, 1, override_msg='Filename should contain at least one \'_\' in the end')
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Function


class TaskMultiFragments(Rule):
    def __init__(self):
        super().__init__(id='oelint.task.multifragments',
                         severity='info',
                         message='Multiple fragments of the same function in the same file should be merged')

    def check(self, _file, stash):
        res = []
        _stash = stash.GetItemsFor(
            filename=_file, classifier=Function.CLASSIFIER)
        _known_matches = []
        for item in _stash:
            _needle = (item.FuncName, item.SubItem, item.Origin)
            if _needle in _known_matches:
                res += self.finding(item.Origin, item.InFileLine)
            _known_matches.append(_needle)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarDescSame(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.descriptionsame',
                         severity='warning',
                         message='\'DESCRIPTION\' is the same a \'SUMMARY\' - it can be removed then')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='DESCRIPTION')
        items_sum = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                      attribute=Variable.ATTR_VAR, attributeValue='SUMMARY')
        for i in items:
            _same = [x for x in items_sum if x.VarValueStripped == i.VarValueStripped]
            if any(_same):
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarHomepagePrefix(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.homepageprefix',
                         severity='warning',
                         message='\'HOMEPAGE\' should start with \'http://\' or \'https://\'')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='HOMEPAGE')
        for i in items:
            if not any([x for x in ['https://', 'http://'] if i.VarValueStripped.strip().startswith(x)]):
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarBbclassextend(Rule):
    def __init__(self):
        super().__init__(id='oelint.var.bbclassextend',
                         severity='info',
                         message='BBCLASSEXTEND should be set if possible')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='BBCLASSEXTEND')
        items_inherit = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue='inherit')
        if not any(items):
            if not any([x for x in items_inherit if x.VarValue.find('native') != -1]):
                res += self.finding(_file, 0)
        return res



import re

from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarMultiInherit(Rule):
    def __init__(self):
        super().__init__(id='oelint.var.multiinherit',
                         severity='warning',
                         message='\'{INH}\' is included multiple times')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='inherit')
        keys = []
        for i in items:
            for y in [x.strip() for x in re.split(r'\s|,|\t|\x1b', i.VarValue) if x]:
                if y not in keys:
                    keys.append(y)
                else:
                    res += self.finding(i.Origin, i.InFileLine,
                                        self.Msg.replace('{INH}', y))
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Function
from oelint_parser.cls_item import TaskAssignment
from oelint_parser.constants import CONSTANTS


class TaskDocStrings(Rule):
    def __init__(self):
        super().__init__(id='oelint.task.docstrings',
                         severity='info',
                         message='Every custom task should have a doc string set by task[doc] = \'\'')

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER):
            if item.FuncName in CONSTANTS.FunctionsKnown or not item.FuncName:
                # Skip for buildin tasks or anonymous python functions
                continue
            if item.IsAppend():
                # In case it's an append operation, there has to be an original
                # so don't raise any warnings here
                continue
            _ta = stash.GetItemsFor(filename=_file, classifier=TaskAssignment.CLASSIFIER,
                                    attribute='FuncName', attributeValue=item.FuncName)
            if not any(_ta):
                res += self.finding(item.Origin, item.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Include


class FileIncludeVsRequire(Rule):
    def __init__(self):
        super().__init__(id='oelint.file.requireinclude',
                         severity='warning',
                         message='Use \'require {FILE}\' instead of \'include {FILE}\'')

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file,
                                      classifier=Include.CLASSIFIER):
            if item.Statement == 'include':
                res += self.finding(item.Origin, item.Line,
                                    self.Msg.replace('{FILE}', item.IncName))
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarNativeSDKFilename(Rule):
    def __init__(self):
        super().__init__(id='oelint.var.nativesdkfilename',
                         severity='warning',
                         message='nativesdk-recipe-files should include \'nativesdk-\' in file name')

    def check(self, _file, stash):
        res = []
        items = [x for x in
                 stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue='inherit')
                 if x.VarValueStripped == 'nativesdk']
        if any(items):
            if _file.find('nativesdk-') == -1:
                res += self.finding(_file, 0)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarDependsAppend(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.dependsappend',
                         severity='error',
                         message='DEPENDS should only be appended, not overwritten')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='DEPENDS')
        for i in items:
            if not i.AppendOperation():
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarNativeFilename(Rule):
    def __init__(self):
        super().__init__(id='oelint.var.nativefilename',
                         severity='warning',
                         message='native-recipe-files should include \'-native\' in file name')

    def check(self, _file, stash):
        res = []
        items = [x for x in
                 stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue='inherit')
                 if x.VarValueStripped == 'native']
        if any(items):
            if _file.find('-native') == -1:
                res += self.finding(_file, 0)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarSectionLowercase(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.doublemodify',
                         severity='error',
                         message='Multiple modifiers of append/prepend/remove/+= found in one operation')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if len(i.AppendOperation()) > 1:
                res += self.finding(i.Origin, i.InFileLine)
        return res



import re

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Function
from oelint_parser.cls_item import Variable
from oelint_parser.constants import CONSTANTS
from oelint_parser.helper_files import get_valid_package_names
from oelint_parser.parser import INLINE_BLOCK


class VarPnBpnUsage(Rule):
    def __init__(self):
        super().__init__(id='oelint.func.specific',
                         severity='error',
                         message='\'{func}\' is set specific to [\'{machine}\'] or [\'{distro}\'], but isn\'t known from PACKAGES, MACHINE, DISTRO, or resources')

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER,
                                  attribute=Function.ATTR_FUNCNAME)
        _comp = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR,
                                  attributeValue='COMPATIBLE_MACHINE')
        _packages = get_valid_package_names(stash, _file)
        _valid_funcs = ['pkg_preinst',
                        'pkg_postinst', 'pkg_prerm', 'pkg_postrm']
        for b in ['pkg_preinst', 'pkg_postinst', 'pkg_prerm', 'pkg_postrm']:
            _valid_funcs += ['{a}-{b}'.format(a=b, b=p)
                             for p in _packages if p.strip() and p != INLINE_BLOCK]
        for i in items:
            _distro = i.GetDistroEntry()
            if _distro in CONSTANTS.DistrosKnown:
                continue

            _machine = i.GetMachineEntry()
            if not _machine or _machine in CONSTANTS.MachinesKnown:
                continue
            if i.FuncName in _valid_funcs:
                continue
            if _machine in ['ptest']:
                # known exceptions
                continue
            if _comp and re.match(''.join(x.VarValueStripped for x in _comp), _machine):
                continue
            res += self.finding(i.Origin, i.InFileLine,
                                override_msg=self.Msg.format(func=i.FuncName, machine=_machine, distro=_distro))
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.helper_files import get_scr_components
from oelint_parser.parser import INLINE_BLOCK


class VarSRCUriOptions(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.srcurioptions',
                         severity='warning',
                         message='<FOO>')
        self._general_options = [
            'apply',
            'destsuffix',
            'name',
            'patchdir',
            'striplevel',
            'subdir',
            'unpack',
        ]
        self._valid_options = {
            'az': [],
            'bzr': [
                'protocol',
                'scmdata',
            ],
            'crcc': [
                'module',
                'proto',
                'vob',
            ],
            'cvs': [
                'date',
                'fullpath',
                'localdir',
                'method',
                'module',
                'norecurse',
                'port',
                'rsh',
                'scmdata',
                'tag',
            ],
            'file': [
                'downloadfilename',
            ],
            'ftp': [
                'downloadfilename',
            ],
            'git': [
                'branch',
                'destsuffix',
                'nobranch',
                'nocheckout',
                'protocol',
                'rebaseable',
                'rev',
                'subdir',
                'subpath',
                'tag',
                'usehead',
                'user',
            ],
            'gitsm': [
                'branch',
                'destsuffix',
                'nobranch',
                'nocheckout',
                'protocol',
                'rebaseable',
                'rev',
                'subpath',
                'tag',
                'usehead',
            ],
            'gitannex': [],
            'hg': [
                'module',
                'rev',
                'scmdata',
                'protocol',
            ],
            'http': [
                'downloadfilename',
            ],
            'https': [
                'downloadfilename',
            ],
            'osc': [
                'module',
                'protocol',
                'rev',
            ],
            'p4': [
                'revision',
            ],
            'repo': [
                'branch',
                'manifest',
                'protocol',
            ],
            'ssh': [],
            's3': [
                'downloadfilename',
            ],
            'sftp': [
                'downloadfilename',
                'protocol',
            ],
            'npm': [
                'name',
                'noverify',
                'version',
            ],
            'npmsw': [
                'dev',
            ],
            'svn': [
                'module',
                'path_spec',
                'protocol',
                'rev',
                'scmdata',
                'ssh',
                'transportuser',
            ],
        }

    def __analyse(self, item, _input, _index):
        _url = get_scr_components(_input)
        res = []
        if 'scheme' not in _url:
            return res  # pragma: no cover
        # For certain types of file:// url parsing fails
        # ignore those
        if _url['scheme'] not in self._valid_options.keys() and not _input.strip().startswith('file://') and _url['scheme']:
            res += self.finding(item.Origin, item.InFileLine + _index,
                                'Fetcher \'{a}\' is not known'.format(a=_url['scheme']))
        else:
            for k, v in _url['options'].items():
                if _url['scheme'] not in self._valid_options:
                    continue  # pragma: no cover
                if k == 'type' and v == 'kmeta':
                    continue  # linux-yocto uses this option to indicate kernel metadata sources
                if k not in self._valid_options[_url['scheme']] + self._general_options:
                    res += self.finding(item.Origin, item.InFileLine + _index,
                                        'Option \'{a}\' is not known with this fetcher type'.format(a=k))
        return res

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')
        for item in items:
            if any([item.Flag.endswith(x) for x in ['md5sum', 'sha256sum']]):
                # These are just the hashes
                continue
            lines = [y.strip('"') for y in item.get_items() if y]
            for x in lines:
                if x == INLINE_BLOCK:
                    continue
                res += self.__analyse(item, x, lines.index(x))
        return res



import re

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class VarMultiLineIndent(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.multilineident',
                         severity='info',
                         message='On a multiline assignment, line indent is desirable. {a} set, {b} desirable')

    def __line_stats(self, raw):
        _map = {}
        _lines = [x for x in re.split(r'\t|\x1b', raw) if x and x.strip()]
        for index, value in enumerate(_lines):
            _map[index] = len(value) - len(value.lstrip())
        _distribution = {x: list(_map.values()).count(x)
                         for x in set(_map.values())}

        if not any(_distribution):
            return (0, list(_map.values()))

        return (max(_distribution, key=lambda x: _distribution[x]), list(_map.values()))

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if not i.IsMultiLine():
                continue
            _likeliest_indent, _indent_map = self.__line_stats(
                i.VarValueStripped)
            _likeliest_indent = max(4, _likeliest_indent)
            for index, value in enumerate(_indent_map):
                if value != _likeliest_indent:
                    res += self.finding(i.Origin, i.InFileLine + index,
                                        self.format_message(a=value, b=_likeliest_indent))
        return res
